from rhoknp.props.dependency import DepType
from rhoknp.props.feature import FeatureDict
from rhoknp.props.memo import MemoTag
from rhoknp.props.named_entity import NamedEntity, NamedEntityCategory
from rhoknp.props.semantics import SemanticsDict

__all__ = [
    "DepType",
    "FeatureDict",
    "MemoTag",
    "NamedEntity",
    "NamedEntityCategory",
    "SemanticsDict",
]
