<!DOCTYPE html>
<html lang="ja">
    <head>
        <meta charset="UTF-8">
        <meta name="keywords" content="nlp,japanese" />
        <meta name="viewport" content="width=device-width, initial-scale=1">
        <title>{{ title }}</title>
        <base href="{{ base_url }}">
        <meta name="description"
              content="A web interface for Japanese text analyzer Juman++, KNP, and KWJA provided by rhoknp">
        <link rel="stylesheet"
              href="https://cdn.jsdelivr.net/npm/bootstrap@5.3.0-alpha3/dist/css/bootstrap.min.css"
              integrity="sha384-KK94CHFLLe+nY2dmCWGMq91rCGa5gtU4mk92HdvYe+M/SXH301p5ILy+dN9+nJOZ"
              crossorigin="anonymous">
        <link rel="stylesheet"
              href="{{ url_for('static', path='/css/style.css') }}">
        <link id="favicon"
              rel="icon"
              type="image/x-icon"
              href="{{ url_for('static', path='/images/favicon.ico') }}">
        <link rel="apple-touch-icon"
              sizes="180x180"
              href="{{ url_for('static', path='/images/apple-touch-icon.png') }}">
        <script src="https://cdn.jsdelivr.net/npm/bootstrap@5.3.0-alpha3/dist/js/bootstrap.bundle.min.js"
                integrity="sha384-ENjdO4Dr2bkBIFxQpeoTz1HIcje39Wm4jDKdf19U8gI4ddQ3GYNS7NTKfAdVQSZe"
                crossorigin="anonymous"
                async></script>
        <script src="{{ url_for('static', path='/js/script.js') }}" async></script>
    </head>
    <body>
        {% include "components/navbar.jinja2" %}
        <div class="container mt-3">
            {% include "components/form.jinja2" %}
            {% if analyzed_document %}
                <hr>
                {% include "components/raw_input.jinja2" %}
                <h6>解析結果</h6>
                <div class="result">
                    {% include "components/show_all_button.jinja2" %}
                    {% include "components/hide_all_button.jinja2" %}
                    {% block result %}
                    {% endblock result %}
                </div>
            {% endif %}
            {% if error %}
                {% include "components/error.jinja2" %}
            {% endif %}
        </div>
    </body>
</html>
