<div class="accordion-item" id="discourse-parsing">
    <h6 class="accordion-header" id="heading-discourse-parsing">
        <button class="accordion-button collapsed"
                type="button"
                data-bs-toggle="collapse"
                data-bs-target="#body-discourse-parsing"
                aria-expanded="true"
                aria-controls="body-discourse-parsing">談話関係解析</button>
    </h6>
    <div id="body-discourse-parsing"
         class="accordion-collapse collapse"
         aria-labelledby="heading-discourse-parsing">
        <div class="accordion-body">
            <table class="table table-sm table-bordered">
                <thead>
                    <tr>
                        <th scope="col">談話関係</th>
                        <th scope="col">タイプ</th>
                        <th scope="col">Modifier</th>
                        <th scope="col">Head</th>
                    </tr>
                </thead>
                <tbody>
                    {% for clause in analyzed_document.clauses %}
                        {% for discourse_relation in clause.discourse_relations %}
                            <tr>
                                <td>{{ discourse_relation.label.value }}</td>
                                <td>
                                    {% if discourse_relation.is_explicit %}
                                        明示的
                                    {% else %}
                                        非明示的
                                    {% endif %}
                                </td>
                                <td>{{ discourse_relation.modifier.text }}</td>
                                <td>{{ discourse_relation.head.text }}</td>
                            </tr>
                        {% endfor %}
                    {% endfor %}
                </tbody>
            </table>
        </div>
    </div>
</div>
