<div class="alert alert-danger mt-4" role="alert">
    <p>解析器の実行中に以下のエラーが発生しました。</p>
    <pre>{{ error }}</pre>
    <p>
        Github の <a href="https://github.com/ku-nlp/rhoknp/issues">Issue</a> に報告してください。
    </p>
</div>
