<form>
    <div>
        <label for="text" class="form-label">テキスト</label>
        <textarea class="form-control" id="text" name="text" rows="3" required>{{ text }}</textarea>
    </div>
    <p class="text-muted text-small">Model: {{ version }}</p>
    <button type="submit" class="btn btn-outline-primary mt-3">解析</button>
</form>
