<div class="accordion-item accordion-item-default-open"
     id="morphological-analysis">
    <h6 class="accordion-header" id="heading-morphological-analysis">
        <button class="accordion-button collapsed"
                type="button"
                data-bs-toggle="collapse"
                data-bs-target="#body-morphological-analysis"
                aria-expanded="true"
                aria-controls="body-morphological-analysis">形態素解析</button>
    </h6>
    <div id="body-morphological-analysis"
         class="accordion-collapse collapse"
         aria-labelledby="heading-morphological-analysis">
        <div class="accordion-body">
            <table class="table table-sm table-bordered">
                <thead>
                    <tr>
                        <th scope="col">表層文字列</th>
                        <th scope="col">読み</th>
                        <th scope="col">原形</th>
                        <th scope="col">品詞</th>
                        <th scope="col">品詞細分類</th>
                        <th scope="col">活用型</th>
                        <th scope="col">活用形</th>
                        <th scope="col">意味情報</th>
                    </tr>
                </thead>
                <tbody>
                    {% for morpheme in analyzed_document.morphemes %}
                        <tr>
                            <td>{{ morpheme.text }}</td>
                            <td>{{ morpheme.reading }}</td>
                            <td>{{ morpheme.lemma }}</td>
                            <td>{{ morpheme.pos }}</td>
                            <td>{{ morpheme.subpos }}</td>
                            <td>{{ morpheme.conjtype }}</td>
                            <td>{{ morpheme.conjform }}</td>
                            <td>{{ morpheme.semantics.to_sstring().strip('"') }}</td>
                        </tr>
                    {% endfor %}
                </tbody>
            </table>
        </div>
    </div>
</div>
