{% extends "base.jinja2" %}
{% block result %}
    <div class="accordion">
        {% include "components/word_splitting.jinja2" %}
        {% include "components/morphological_analysis.jinja2" %}
        {% with raw_output = analyzed_document.to_jumanpp() %}
            {% include "components/raw_output.jinja2" %}
        {% endwith %}
    </div>
{% endblock result %}
