<div class="accordion-item" id="raw-output">
    <h6 class="accordion-header" id="heading-raw-output">
        <button class="accordion-button collapsed"
                type="button"
                data-bs-toggle="collapse"
                data-bs-target="#body-raw-output"
                aria-expanded="true"
                aria-controls="body-raw-output">生の解析結果</button>
    </h6>
    <div id="body-raw-output"
         class="accordion-collapse collapse"
         aria-labelledby="heading-raw-output">
        <div class="accordion-body">
            <pre>{{ raw_output }}</pre>
        </div>
    </div>
</div>
