<div class="accordion-item accordion-item-default-open"
     id="dependency-parsing">
    <h6 class="accordion-header" id="heading-dependency-parsing">
        <button class="accordion-button collapsed"
                type="button"
                data-bs-toggle="collapse"
                data-bs-target="#body-dependency-parsing"
                aria-expanded="true"
                aria-controls="body-dependency-parsing">構文解析・述語項構造解析</button>
    </h6>
    <div id="body-dependency-parsing"
         class="accordion-collapse collapse"
         aria-labelledby="heading-dependency-parsing">
        <div class="accordion-body">
            <pre>{{ tree }}</pre>
        </div>
    </div>
</div>
